ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c---------------------------------------------------------------------c
c                                                                     c
c This program is used to convert the old file, created by calling    c
c apndgrd multiple times while looping through pressure levels, into  c
c a three dimensional cylindrical grid (with coordinates of azimuth,  c
c distance, and pressure).  Distance is measured along a great circle c
c arc originating from the storm center. Each grid value is inter-    c
c polated from the old file to the new using bilinear interpolation.  c
c                                                                     c
c---------------------------------------------------------------------c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      PROGRAM rec2cyl

      IMPLICIT NONE

      INTEGER, PARAMETER :: iunit_gr_out=21, iunit_gr_in=22, 
     & iunit_gr_rect=23, iunit_gr_cyl=24

      REAL ivals(20), ovals(20), rect_header(20), cyl_header(10),
     & extargs(10)
      REAL, ALLOCATABLE :: grid_data(:)
      REAL, ALLOCATABLE :: rect_data(:,:,:)
      REAL, ALLOCATABLE :: cyl_data(:,:,:)
      INTEGER nxdim, nydim, nsiz
      REAL slat, wlon, lat_inc, lon_inc
      REAL lato, lono, range, startaz, endaz, botlev, toplev
      INTEGER nrange, nazimuth, nlevs
      INTEGER i

c  Open the files

      CALL open_files(iunit_gr_out,iunit_gr_in,
     &                iunit_gr_rect,iunit_gr_cyl)

c  Load the grid header information from the grads output file

      READ(iunit_gr_out) ovals

c  Load extended function arguments from file

      CALL load_extanded_args(extargs,10)
      
      lato     = extargs(1)
      lono     = extargs(2)
      nazimuth = nint(extargs(3))
      nrange   = nint(extargs(4))
      nlevs    = nint(extargs(5))
      startaz  = extargs(6)
      endaz    = extargs(7)
      range    = extargs(8)
      botlev   = extargs(9)
      toplev   = extargs(10)
    
c  Loat the header information from the rectangular grid file

      READ(iunit_gr_rect) rect_header
      REWIND(iunit_gr_rect)

      nxdim  = nint(rect_header(4))
      nydim  = nint(rect_header(5))
      wlon   = rect_header(8)
      lon_inc = rect_header(9)
      slat   = rect_header(10)
      lat_inc = rect_header(11)

c  Allocate appropriatly sized grids for the rectangular and
c  cylindrical data

      ALLOCATE(rect_data(nxdim,nydim,nlevs))
      ALLOCATE(cyl_data(nazimuth,nrange,nlevs))
      
      CALL read_rect_file(iunit_gr_rect,rect_data,nxdim,nydim,nlevs)

      CALL intrp_rect2cyl(rect_data,nxdim,nydim,nlevs,
     &                    cyl_data,nazimuth,nrange,
     &                    lato,lono,startaz,endaz,range,
     &                    slat,wlon,lat_inc,lon_inc)

c  Create file headera and write cylindrical grid data to file

      cyl_header(1)  = lato
      cyl_header(2)  = lono
      cyl_header(3)  = nazimuth + 0.0
      cyl_header(4)  = nrange + 0.0
      cyl_header(5)  = nlevs + 0.0
      cyl_header(6)  = startaz
      cyl_header(7)  = endaz
      cyl_header(8)  = range
      cyl_header(9)  = botlev
      cyl_header(10) = toplev
     
      WRITE(iunit_gr_cyl) cyl_header
      WRITE(iunit_gr_cyl) cyl_data

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c  This section just prepares dummy data to write to the grads
c  input file.

c  Get the essential dimension informatio from the header and allocate
c  memory for the grid data

      nxdim = nint(ovals(4))
      nydim = nint(ovals(5))
      nsiz  = nxdim * nydim

      ALLOCATE(grid_data(nsiz))

c  Store grid data into memory

      READ(iunit_gr_out) grid_data

c  Copy grid header from grads output to grid header for GrADS input

      DO i=1,20
        ivals(i) = ovals(i)
      ENDDO

c  Write out dummy data to function result file

      CALL write_frf(iunit_gr_in,ivals,grid_data,nsiz)

      DEALLOCATE(grid_data)

c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

c Deallocata all arrays and close all open files

      DEALLOCATE(rect_data)
      DEALLOCATE(cyl_data)

      CALL close_files(iunit_gr_out,iunit_gr_in,
     &                 iunit_gr_rect,iunit_gr_cyl)

      END

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c---------------------------------------------------------------------c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      SUBROUTINE open_files(iunit_gr_out,iunit_gr_in,
     &                      iunit_gr_rect,iunit_gr_cyl)
  
      IMPLICIT NONE

      INTEGER iunit_gr_out, iunit_gr_in, iunit_gr_rect, iunit_gr_cyl
      INTEGER nargs, dtffrmt,i,j
      REAL vals(20)
      CHARACTER(LEN=32) ifile
      CHARACTER(LEN=32) ofile
      CHARACTER(LEN=41) dirfile

c  Open data transfer files

      OPEN (iunit_gr_out,FILE='./udfs/udf.rec2cyl.out',
     &      STATUS='unknown', FORM='unformatted',ACCESS='sequential',
     &      POSITION='rewind',ERR=910)
           
      OPEN (iunit_gr_in,FILE='./udfs/udf.rec2cyl.in',STATUS='unknown',
     &      FORM='unformatted',ACCESS='sequential',ERR=920)

      GOTO 400

  910 PRINT *, '!!! ERROR: unable to open udf.rec2cyl.out'
      STOP 50

  920 PRINT *, '!!! ERROR: unable to open udf.rec2cyl.in'
      STOP 50

  400 CONTINUE

c  Read data transfer file header

      READ(iunit_gr_out,err=915) vals

c  Make sure number of arguments is 3

      nargs = nint(vals(1))
      IF (nargs /= 3) THEN
        PRINT *, '!!! ERROR: incorrect number of arguments'
        STOP 50
      ENDIF
      
c  Check to make sure the data transfer file is the correct format
c  *** NOTE: TPM commented this out on 11/29/2006 because the 1.9
c      version of grads was putting non-floating point values into
c      this array, core-dumping the program.

c      dtffrmt = nint(vals(2))
c      IF (dtffrmt /= 0) THEN
c        PRINT *, '!!! ERROR: incorrect file transfer format in rec2cyl'
c        do j = 1,20
c          print *,'BBB j= ',j,' vals(j)= ',vals(j)
c          write (51,401) j,vals(j)
c        enddo
c	STOP 50
c 401    format (1x,'i= ',i4,' vals(i)= ',f15.5)
c      ENDIF

c  Read the name of the rectangular grid file

      READ(iunit_gr_out) ifile
      dirfile = './udfs/'//ifile

c  Open the specified file

      OPEN (iunit_gr_rect,FILE=dirfile,STATUS='old',
     &      FORM='unformatted',ACCESS='sequential',
     &      POSITION='rewind',ERR=930)
     
c  Read the name of the cylindrical grid file

      READ(iunit_gr_out) ofile
      dirfile = './udfs/'//ofile

c  Open the specified file

      OPEN (iunit_gr_cyl,FILE=dirfile,STATUS='replace',
     &      FORM='unformatted',ACCESS='sequential',
     &      ERR=940)

      GOTO 500

  915 print *,'!!! ERROR READING vals array in rec2cyl'
      stop 50

  930 PRINT *, '!!! ERROR: unable to open '//ifile
      STOP 50
      
  940 PRINT *, '!!! ERROR: unable to open '//ofile

  500 CONTINUE

      RETURN
      END

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c---------------------------------------------------------------------c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      SUBROUTINE write_frf(iunit_gr_in,ivals,grid_data,nsiz)

c  This just writes dummy data to the function result file

      IMPLICIT NONE

      INTEGER iunit_gr_in
      INTEGER nsiz
      REAL vals(20), ivals(20), grid_data(nsiz)

      vals(1) = 0.0
      vals(2) = 0.0

      WRITE(iunit_gr_in,err=910) vals
      WRITE(iunit_gr_in,err=920) ivals
      WRITE(iunit_gr_in,err=930) grid_data

      goto 500

  910 print *,'!!! ERROR WRITING vals array in write_frf'
      stop 50
  
  920 print *,'!!! ERROR WRITING ivals array in write_frf'
      stop 50
  
  930 print *,'!!! ERROR WRITING grid_data array in write_frf'
      stop 50

  500 continue

      RETURN
      END

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c---------------------------------------------------------------------c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      SUBROUTINE load_extanded_args(extargs,nargs)

c  This loads the extra arguments that wouldn't 'fit' in the grads
c  function data transfer file

      IMPLICIT NONE
      
      INTEGER, PARAMETER :: iunit_args = 54
      INTEGER nargs, i
      REAL extargs(nargs)

      OPEN(iunit_args,FILE='./udfs/extended_udf_args.txt',STATUS='OLD',
     &     FORM='formatted',ACCESS='sequential',POSITION='rewind',
     &     ERR=910)

      GOTO 500

  910 PRINT *, 'error opening xtended_udf_args.txt'

  500 CONTINUE

      DO i=1,nargs
        READ(iunit_args, 210) extargs(i)
      ENDDO

  210 format(f20.10)

      RETURN
      END


ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c---------------------------------------------------------------------c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      SUBROUTINE read_rect_file(iunit_gr_rect,rect_data,
     &                          nxdim,nydim,nlevs)

c  Transfer the grid data from file to 3 dimensional array in memory

      IMPLICIT NONE

      INTEGER iunit_gr_rect, nxdim, nydim, nlevs
      REAL rect_data(nxdim,nydim,nlevs)
      REAL, ALLOCATABLE :: buffer(:,:)
      REAL rect_header(20)
      INTEGER x, y, z

      ALLOCATE(buffer(nxdim,nydim))

      DO z=1,nlevs
        READ(iunit_gr_rect) rect_header
        READ(iunit_gr_rect) buffer
        DO y=1,nydim
          DO x=1,nxdim
            rect_data(x,y,z) = buffer(x,y)
          ENDDO
        ENDDO
      ENDDO

      DEALLOCATE(buffer)
      
      RETURN
      END

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c---------------------------------------------------------------------c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      SUBROUTINE intrp_rect2cyl(rect_data,nxdim,nydim,nlevs,
     &                          cyl_data,nazimuth,nrange,
     &                          lato,lono,startaz,endaz,range,
     &                          slat,wlon,lat_inc,lon_inc)

c  This subroutine interpolates a 3 dimensional (x,y,z) grid array over
c  a rectangular region into a 3 dimensional (s,r,z) grid array over
c  a cylindrical region
      
      IMPLICIT NONE
      
      INTEGER nxdim, nydim, nlevs, nazimuth, nrange
      REAL rect_data(nxdim,nydim,nlevs)
      REAL cyl_data(nazimuth,nrange,nlevs)
      REAL lato, lono, lono2, startaz, endaz, range
      REAL slat, wlon, lat_inc, lon_inc

      INTEGER i, j, z
      REAL s, s1, s2, r, s_inc, r_inc
      INTEGER xo, yo
      REAL lat, lon, x, y, delx, dely
      REAL value

      REAL cos_lato, sin_lato, cos_lono, sin_lono
      REAL cos_dist, sin_dist, cos_bear, sin_bear

      REAL, PARAMETER :: radpdeg = 0.017453293

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c  Here we interpolat the data at center of the cylinder (r=0) at
c  each level

      x    = (lono - wlon) / lon_inc + 1.0
      y    = (lato - slat) / lat_inc + 1.0
      xo   = floor(x)
      yo   = floor(y)
      delx = x - xo
      dely = y - yo

      DO z=1,nlevs
        value = (1.0-delx)*(1.0-dely)*rect_data(xo,yo,z)
        value = value + (delx)*(1.0-dely)*rect_data(xo+1,yo,z)
        value = value + (1.0-delx)*(dely)*rect_data(xo,yo+1,z)
        value = value + (delx)*(dely)*rect_data(xo+1,yo+1,z)
        DO i=1,nazimuth
          cyl_data(i,1,z) = value
        ENDDO
      ENDDO
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

c  Add a multiple of 360 to the ending azimuth to make it greater than 
c  the starting azimuth (this makes for easier looping)

      s1 = startaz
      s2 = endaz
      DO WHILE (s2 <= s1)
        s2 = s2 + 360.0
      ENDDO

c  Here we force the starting longitude to be <= 180 and > -180
c  This is to make sure the values are compatible with the db2llfast
c  subroutine.  Negative values should be west and positive vales east
c  for the db2llfast subroutine.

      lono2 = lono
      DO WHILE (lono2 < -180.0)
        lono2 = lono2 + 360.0
      ENDDO
      DO WHILE (lono2 >= 180.0)
        lono2 = lono2 - 360.0
      ENDDO

c  Calculate increments of radius(r) and azimuth(s)

      r_inc = range/(nrange-1)
      s_inc = (s2-s1)/(nazimuth-1)

c  Precalculate sins and cosines of unvarying variables so we don't
c  have to in the inner loops

      cos_lato = cos(lato*radpdeg)
      sin_lato = sin(lato*radpdeg)
      cos_lono = cos(lono2*radpdeg)
      sin_lono = sin(lono2*radpdeg)

c  Below is the loop that interpolates all the data at r > 0.  Each 
c  loop goes around in a clockwise circle from the starting asimuth to 
c  the ending azimuth interpolating to nazimuth locations at nlevs 
c  different z levels.

      r = r_inc      

      DO i=2,nrange
c        PRINT *, 'RAD= ',r
        
        cos_dist = cos(r*radpdeg)
        sin_dist = sin(r*radpdeg)

        s = s1

        DO j=1,nazimuth

          cos_bear = cos(s*radpdeg)
          sin_bear = sin(s*radpdeg)

c  Get the latitude and longitude of the point on the circle

          CALL db2llfast(cos_lato,sin_lato,cos_lono,sin_lono,
     &                   cos_dist,sin_dist,cos_bear,sin_bear,
     &                   lon, lat)

c  Convert the longitude from a value between -180 and 180 back into
c  GrADS world coordinates

          lon = lon + lono - lono2

c  Do the interpolation

          x    = (lon - wlon) / lon_inc + 1.0
          y    = (lat - slat) / lat_inc + 1.0
          xo   = floor(x)
          yo   = floor(y)
          delx = x - xo
          dely = y - yo
c          PRINT *,'AZ= ',s,'lat= ',lat,'lon= ',lon

          DO z=1,nlevs
            value = (1.0-delx)*(1.0-dely)*rect_data(xo,yo,z)
            value = value + (delx)*(1.0-dely)*rect_data(xo+1,yo,z)
            value = value + (1.0-delx)*(dely)*rect_data(xo,yo+1,z)
            value = value + (delx)*(dely)*rect_data(xo+1,yo+1,z)
            cyl_data(j,i,z) = value
c            PRINT *,'z= ',z,'value= ',value
          ENDDO

          s = s + s_inc

        ENDDO
        
        r = r + r_inc
        
      ENDDO

      RETURN
      END

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c---------------------------------------------------------------------c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      SUBROUTINE db2llfast(cos_lato,sin_lato,cos_lono,sin_lono,
     &                     cos_dist,sin_dist,cos_bear,sin_bear,
     &                     lont, latt)

      IMPLICIT NONE

      REAL cos_lato, sin_lato, cos_lono, sin_lono, cos_dist, sin_dist
      REAL cos_bear, sin_bear, lont, latt

      REAL, PARAMETER :: radpdeg = 0.017453293
      REAL z, y, x, r

      z = cos_dist*sin_lato + cos_lato*sin_dist*cos_bear

      y = cos_lato*cos_lono*cos_dist + sin_dist*
     &    (-sin_lono*sin_bear - sin_lato*cos_lono*cos_bear)

      x = -cos_lato*sin_lono*cos_dist - sin_dist*
     &    (cos_lono*sin_bear - sin_lato*sin_lono*cos_bear)

      r = sqrt(x*x + y*y)

      latt = atan2(z,r) / radpdeg
      lont = -atan2(x,y) / radpdeg
      
      RETURN
      END

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c---------------------------------------------------------------------c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc


      SUBROUTINE close_files(iunit_gr_out,iunit_gr_in,
     &                       iunit_gr_rect, iunit_gr_cyl)

c  Close all files opened in program

      IMPLICIT NONE

      INTEGER iunit_gr_out, iunit_gr_in, iunit_gr_rect, iunit_gr_cyl
      
      CLOSE(iunit_gr_out)
      CLOSE(iunit_gr_in)
      CLOSE(iunit_gr_rect)
      CLOSE(iunit_gr_cyl)

      RETURN
      END

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c---------------------------------------------------------------------c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
